>>>
POST / HTTP/1.0
Content-Length: 12
Content-Type: application/x-www-form-urlencoded
User-Agent: Mozilla


<<<
HTTP/1.0 400 Bad Request
Date: Mon, 31 Aug 2009 20:25:50 GMT
Server: Apache
Connection: close
Content-Type: text/html
Content-Length: 11

Bad Request
>>>
p=0123456789