>>>
GET
<<<
HTTP/1.0 408 Request Time-out
Server: AkamaiGHost
Mime-Version: 1.0
Date: Fri, 27 Sep 2013 16:37:37 GMT
Content-Type: text/html
Content-Length: 218
Expires: Fri, 27 Sep 2013 16:37:37 GMT

<HTML><HEAD>
<TITLE>Request Timeout</TITLE>
</HEAD><BODY>
<H1>Request Timeout</H1>
The server timed out while waiting for the browser's request.<P>
Reference&#32;&#35;2&#46;9efcd4d9&#46;1380708056&#46;0
</BODY></HTML>
