>>>
POST /?p=1&q=2 HTTP/1.0
Content-Type: application/x-www-form-urlencoded
Content-Length: 11

p=3&q=4&z=5
<<<
HTTP/1.0 200 OK
Date: Mon, 31 Aug 2009 20:25:50 GMT
Server: Apache
Connection: close
Content-Type: text/html
Content-Length: 12

Hello World!