>>>
GET /?p=%20 HTTP/1.0
User-Agent: Mozilla


<<<
HTTP/1.0 200 OK
Date:
<<<
 Mon,
<<<
 31 Aug 2009 20:25:50 GMT
Server:
<<<
 Apache
Connection: close
Content-Type: text/html
Transfer-Encoding: chunked

b
Hello World
1
!
0
Set-Cookie:
<<<
 name=
<<<
value
Another-Header:
<<<
 Header-Value

