>>>
CONNECT www.feistyduck.com:80 HTTP/1.1
Host: www.feistyduck.com

HEAD / HTTP/1.0


<<<
HTTP/1.1 301 Moved Permanently
Date: Wed, 06 Jan 2010 17:41:34 GMT
Server: Apache
Location: https://www.feistyduck.com/
Vary: Accept-Encoding
Content-Length: 235
Content-Type: text/html; charset=iso-8859-1

<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
<html><head>
<title>301 Moved Permanently</title>
</head><body>
<h1>Moved Permanently</h1>
<p>The document has moved <a href="https://www.feistyduck.com/">here</a>.</p>
</body></html>

HTTP/1.1 301 Moved Permanently
Date: Wed, 06 Jan 2010 17:41:46 GMT
Server: Apache
Location: https://www.feistyduck.com/
Vary: Accept-Encoding
Connection: close
Content-Type: text/html; charset=iso-8859-1

