>>>
GET
>>>
 /?x=y
>>>
 HTTP/1.0
User-Agent:
>>>
 Test
>>>
 User
>>>
 Agent
Host: www.example.com


<<<
HTTP/1.0
<<<
 200
<<<
 OK
Date:
<<<
 Mon,
<<<
 31
<<<
 Aug
<<<
 2009
<<<
 20:25:50
<<<
 GMT
Server: Apache
Connection: close
Content-Type: text/html
Content-Length: 12

Hello World!