>>>
GET /blaaaaaaaaaaaaaaaaaaaaaaaaa
<<<
HTTP/1.0 414 Request-URI Too Large
Server: MyBigFatServer
Mime-Version: 1.0
Date: Fri, 27 Sep 2013 16:37:37 GMT
Content-Type: text/html
Content-Length: 139
Expires: Fri, 27 Sep 2013 16:37:37 GMT

<HTML><HEAD>
<TITLE>Request-URI Too Large</TITLE>
</HEAD><BODY>
<H1>Request-URI Too Large</H1>
The Request-URI is Too Large
</BODY></HTML>
