>>>
GET / HTTP/1.1
Host: www.example.com
Authorization: Digest username = ivanr, realm="Book Review",
 nonce="OgmPjb/jAwA=7c5a49c2ed9416dba1b04b5307d6d935f74a859d",
 uri="/review/", algorithm=MD5, response="3c430d26043cc306e0282635929d57cb",
 qop=auth, nc=00000004, cnonce="c3bcee9534c051a0"

